# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# [start workflow_declare]
r"""
A example workflow for task switch.

This example will create four task in single workflow, with three shell task and one switch task. Task switch
have one upstream which we declare explicit with syntax `parent >> switch`, and two downstream automatically
set dependence by switch task by passing parameter `condition`. The graph of this workflow like:
                      --> switch_child_1
                    /
parent -> switch ->
                    \
                      --> switch_child_2
.
"""

from pydolphinscheduler.core.process_definition import ProcessDefinition
from pydolphinscheduler.tasks.shell import Shell
from pydolphinscheduler.tasks.switch import Branch, Default, Switch, SwitchCondition

with ProcessDefinition(
    name="task_switch_example", tenant="tenant_exists", param={"var": "1"}
) as pd:
    parent = Shell(name="parent", command="echo parent")
    switch_child_1 = Shell(name="switch_child_1", command="echo switch_child_1")
    switch_child_2 = Shell(name="switch_child_2", command="echo switch_child_2")
    switch_condition = SwitchCondition(
        Branch(condition="${var} > 1", task=switch_child_1),
        Default(task=switch_child_2),
    )

    switch = Switch(name="switch", condition=switch_condition)
    parent >> switch
    pd.submit()
# [end workflow_declare]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Task datax."""

from typing import Dict, List, Optional

from pydolphinscheduler.constants import TaskType
from pydolphinscheduler.core.database import Database
from pydolphinscheduler.core.task import Task


class CustomDataX(Task):
    """Task CustomDatax object, declare behavior for custom DataX task to dolphinscheduler.

    You provider json template for DataX, it can synchronize data according to the template you provided.
    """

    CUSTOM_CONFIG = 1

    _task_custom_attr = {"custom_config", "json", "xms", "xmx"}

    def __init__(
        self,
        name: str,
        json: str,
        xms: Optional[int] = 1,
        xmx: Optional[int] = 1,
        *args,
        **kwargs
    ):
        super().__init__(name, TaskType.DATAX, *args, **kwargs)
        self.custom_config = self.CUSTOM_CONFIG
        self.json = json
        self.xms = xms
        self.xmx = xmx


class DataX(Task):
    """Task DataX object, declare behavior for DataX task to dolphinscheduler.

    It should run database datax job in multiply sql link engine, such as:
    - MySQL
    - Oracle
    - Postgresql
    - SQLServer
    You provider datasource_name and datatarget_name contain connection information, it decisions which
    database type and database instance would synchronous data.
    """

    CUSTOM_CONFIG = 0

    _task_custom_attr = {
        "custom_config",
        "sql",
        "target_table",
        "job_speed_byte",
        "job_speed_record",
        "pre_statements",
        "post_statements",
        "xms",
        "xmx",
    }

    def __init__(
        self,
        name: str,
        datasource_name: str,
        datatarget_name: str,
        sql: str,
        target_table: str,
        job_speed_byte: Optional[int] = 0,
        job_speed_record: Optional[int] = 1000,
        pre_statements: Optional[List[str]] = None,
        post_statements: Optional[List[str]] = None,
        xms: Optional[int] = 1,
        xmx: Optional[int] = 1,
        *args,
        **kwargs
    ):
        super().__init__(name, TaskType.DATAX, *args, **kwargs)
        self.sql = sql
        self.custom_config = self.CUSTOM_CONFIG
        self.datasource_name = datasource_name
        self.datatarget_name = datatarget_name
        self.target_table = target_table
        self.job_speed_byte = job_speed_byte
        self.job_speed_record = job_speed_record
        self.pre_statements = pre_statements or []
        self.post_statements = post_statements or []
        self.xms = xms
        self.xmx = xmx

    @property
    def task_params(self, camel_attr: bool = True, custom_attr: set = None) -> Dict:
        """Override Task.task_params for datax task.

        datax task have some specials attribute for task_params, and is odd if we
        directly set as python property, so we Override Task.task_params here.
        """
        params = super().task_params
        datasource = Database(self.datasource_name, "dsType", "dataSource")
        params.update(datasource)

        datatarget = Database(self.datatarget_name, "dtType", "dataTarget")
        params.update(datatarget)
        return params



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Module database."""

from typing import Dict

from py4j.protocol import Py4JJavaError

from pydolphinscheduler.exceptions import PyDSParamException
from pydolphinscheduler.java_gateway import launch_gateway


class Database(dict):
    """database object, get information about database.

    You provider database_name contain connection information, it decisions which
    database type and database instance would run task.
    """

    def __init__(self, database_name: str, type_key, database_key, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._database = {}
        self.database_name = database_name
        self[type_key] = self.database_type
        self[database_key] = self.database_id

    @property
    def database_type(self) -> str:
        """Get database type from java gateway, a wrapper for :func:`get_database_info`."""
        return self.get_database_info(self.database_name).get("type")

    @property
    def database_id(self) -> str:
        """Get database id from java gateway, a wrapper for :func:`get_database_info`."""
        return self.get_database_info(self.database_name).get("id")

    def get_database_info(self, name) -> Dict:
        """Get database info from java gateway, contains database id, type, name."""
        if self._database:
            return self._database
        else:
            gateway = launch_gateway()
            try:
                self._database = gateway.entry_point.getDatasourceInfo(name)
            # Handler database source do not exists error, for now we just terminate the process.
            except Py4JJavaError as ex:
                raise PyDSParamException(str(ex.java_exception))
            return self._database



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Task dependent."""

from typing import Dict, Optional, Tuple

from pydolphinscheduler.constants import TaskType
from pydolphinscheduler.core.base import Base
from pydolphinscheduler.core.task import Task
from pydolphinscheduler.exceptions import PyDSJavaGatewayException, PyDSParamException
from pydolphinscheduler.java_gateway import launch_gateway

DEPENDENT_ALL_TASK_IN_WORKFLOW = "0"


class DependentDate(str):
    """Constant of Dependent date value.

    These values set according to Java server side, if you want to add and change it,
    please change Java server side first.
    """

    # TODO Maybe we should add parent level to DependentDate for easy to use, such as
    # DependentDate.MONTH.THIS_MONTH

    # Hour
    CURRENT_HOUR = "currentHour"
    LAST_ONE_HOUR = "last1Hour"
    LAST_TWO_HOURS = "last2Hours"
    LAST_THREE_HOURS = "last3Hours"
    LAST_TWENTY_FOUR_HOURS = "last24Hours"

    # Day
    TODAY = "today"
    LAST_ONE_DAYS = "last1Days"
    LAST_TWO_DAYS = "last2Days"
    LAST_THREE_DAYS = "last3Days"
    LAST_SEVEN_DAYS = "last7Days"

    # Week
    THIS_WEEK = "thisWeek"
    LAST_WEEK = "lastWeek"
    LAST_MONDAY = "lastMonday"
    LAST_TUESDAY = "lastTuesday"
    LAST_WEDNESDAY = "lastWednesday"
    LAST_THURSDAY = "lastThursday"
    LAST_FRIDAY = "lastFriday"
    LAST_SATURDAY = "lastSaturday"
    LAST_SUNDAY = "lastSunday"

    # Month
    THIS_MONTH = "thisMonth"
    LAST_MONTH = "lastMonth"
    LAST_MONTH_BEGIN = "lastMonthBegin"
    LAST_MONTH_END = "lastMonthEnd"


class DependentItem(Base):
    """Dependent item object, minimal unit for task dependent.

    It declare which project, process_definition, task are dependent to this task.
    """

    _DEFINE_ATTR = {
        "project_code",
        "definition_code",
        "dep_task_code",
        "cycle",
        "date_value",
    }

    # TODO maybe we should conside overwrite operator `and` and `or` for DependentItem to
    #  support more easy way to set relation
    def __init__(
        self,
        project_name: str,
        process_definition_name: str,
        dependent_task_name: Optional[str] = DEPENDENT_ALL_TASK_IN_WORKFLOW,
        dependent_date: Optional[DependentDate] = DependentDate.TODAY,
    ):
        obj_name = f"{project_name}.{process_definition_name}.{dependent_task_name}.{dependent_date}"
        super().__init__(obj_name)
        self.project_name = project_name
        self.process_definition_name = process_definition_name
        self.dependent_task_name = dependent_task_name
        if dependent_date is None:
            raise PyDSParamException(
                "Parameter dependent_date must provider by got None."
            )
        else:
            self.dependent_date = dependent_date
        self._code = {}

    def __repr__(self) -> str:
        return "depend_item_list"

    @property
    def project_code(self) -> str:
        """Get dependent project code."""
        return self.get_code_from_gateway().get("projectCode")

    @property
    def definition_code(self) -> str:
        """Get dependent definition code."""
        return self.get_code_from_gateway().get("processDefinitionCode")

    @property
    def dep_task_code(self) -> str:
        """Get dependent tasks code list."""
        if self.is_all_task:
            return DEPENDENT_ALL_TASK_IN_WORKFLOW
        else:
            return self.get_code_from_gateway().get("taskDefinitionCode")

    # TODO Maybe we should get cycle from dependent date class.
    @property
    def cycle(self) -> str:
        """Get dependent cycle."""
        if "Hour" in self.dependent_date:
            return "hour"
        elif self.dependent_date == "today" or "Days" in self.dependent_date:
            return "day"
        elif "Month" in self.dependent_date:
            return "month"
        else:
            return "week"

    @property
    def date_value(self) -> str:
        """Get dependent date."""
        return self.dependent_date

    @property
    def is_all_task(self) -> bool:
        """Check whether dependent all tasks or not."""
        return self.dependent_task_name == DEPENDENT_ALL_TASK_IN_WORKFLOW

    @property
    def code_parameter(self) -> Tuple:
        """Get name info parameter to query code."""
        param = (
            self.project_name,
            self.process_definition_name,
            self.dependent_task_name if not self.is_all_task else None,
        )
        return param

    def get_code_from_gateway(self) -> Dict:
        """Get project, definition, task code from given parameter."""
        if self._code:
            return self._code
        else:
            gateway = launch_gateway()
            try:
                self._code = gateway.entry_point.getDependentInfo(*self.code_parameter)
                return self._code
            except Exception:
                raise PyDSJavaGatewayException("Function get_code_from_gateway error.")


class DependentOperator(Base):
    """Set DependentItem or dependItemList with specific operator."""

    _DEFINE_ATTR = {
        "relation",
    }

    def __init__(self, *args):
        super().__init__(self.__class__.__name__)
        self.args = args

    def __repr__(self) -> str:
        return "depend_task_list"

    @classmethod
    def operator_name(cls) -> str:
        """Get operator name in different class."""
        return cls.__name__.upper()

    @property
    def relation(self) -> str:
        """Get operator name in different class, for function :func:`get_define`."""
        return self.operator_name()

    def set_define_attr(self) -> str:
        """Set attribute to function :func:`get_define`.

        It is a wrapper for both `And` and `Or` operator.
        """
        result = []
        attr = None
        for dependent in self.args:
            if isinstance(dependent, (DependentItem, DependentOperator)):
                if attr is None:
                    attr = repr(dependent)
                elif repr(dependent) != attr:
                    raise PyDSParamException(
                        "Dependent %s operator parameter only support same type.",
                        self.relation,
                    )
            else:
                raise PyDSParamException(
                    "Dependent %s operator parameter support DependentItem and "
                    "DependentOperator but got %s.",
                    (self.relation, type(dependent)),
                )
            result.append(dependent.get_define())
        setattr(self, attr, result)
        return attr

    def get_define(self, camel_attr=True) -> Dict:
        """Overwrite Base.get_define to get task dependent specific get define."""
        attr = self.set_define_attr()
        dependent_define_attr = self._DEFINE_ATTR.union({attr})
        return super().get_define_custom(
            camel_attr=True, custom_attr=dependent_define_attr
        )


class And(DependentOperator):
    """Operator And for task dependent.

    It could accept both :class:`DependentItem` and children of :class:`DependentOperator`,
    and set AND condition to those args.
    """

    def __init__(self, *args):
        super().__init__(*args)


class Or(DependentOperator):
    """Operator Or for task dependent.

    It could accept both :class:`DependentItem` and children of :class:`DependentOperator`,
    and set OR condition to those args.
    """

    def __init__(self, *args):
        super().__init__(*args)


class Dependent(Task):
    """Task dependent object, declare behavior for dependent task to dolphinscheduler."""

    def __init__(self, name: str, dependence: DependentOperator, *args, **kwargs):
        super().__init__(name, TaskType.DEPENDENT, *args, **kwargs)
        self.dependence = dependence

    @property
    def task_params(self, camel_attr: bool = True, custom_attr: set = None) -> Dict:
        """Override Task.task_params for dependent task.

        Dependent task have some specials attribute `dependence`, and in most of the task
        this attribute is None and use empty dict `{}` as default value. We do not use class
        attribute `_task_custom_attr` due to avoid attribute cover.
        """
        params = super().task_params
        params["dependence"] = self.dependence.get_define()
        return params



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Task shell."""

from typing import Optional

from pydolphinscheduler.constants import TaskType
from pydolphinscheduler.core.task import Task
from pydolphinscheduler.exceptions import PyDSParamException


class HttpMethod:
    """Constant of HTTP method."""

    GET = "GET"
    POST = "POST"
    HEAD = "HEAD"
    PUT = "PUT"
    DELETE = "DELETE"


class HttpCheckCondition:
    """Constant of HTTP check condition.

    For now it contain four value:
    - STATUS_CODE_DEFAULT: when http response code equal to 200, mark as success.
    - STATUS_CODE_CUSTOM: when http response code equal to the code user define, mark as success.
    - BODY_CONTAINS: when http response body contain text user define, mark as success.
    - BODY_NOT_CONTAINS: when http response body do not contain text user define, mark as success.
    """

    STATUS_CODE_DEFAULT = "STATUS_CODE_DEFAULT"
    STATUS_CODE_CUSTOM = "STATUS_CODE_CUSTOM"
    BODY_CONTAINS = "BODY_CONTAINS"
    BODY_NOT_CONTAINS = "BODY_NOT_CONTAINS"


class Http(Task):
    """Task HTTP object, declare behavior for HTTP task to dolphinscheduler."""

    _task_custom_attr = {
        "url",
        "http_method",
        "http_params",
        "http_check_condition",
        "condition",
        "connect_timeout",
        "socket_timeout",
    }

    def __init__(
        self,
        name: str,
        url: str,
        http_method: Optional[str] = HttpMethod.GET,
        http_params: Optional[str] = None,
        http_check_condition: Optional[str] = HttpCheckCondition.STATUS_CODE_DEFAULT,
        condition: Optional[str] = None,
        connect_timeout: Optional[int] = 60000,
        socket_timeout: Optional[int] = 60000,
        *args,
        **kwargs
    ):
        super().__init__(name, TaskType.HTTP, *args, **kwargs)
        self.url = url
        if not hasattr(HttpMethod, http_method):
            raise PyDSParamException(
                "Parameter http_method %s not support.", http_method
            )
        self.http_method = http_method
        self.http_params = http_params or []
        if not hasattr(HttpCheckCondition, http_check_condition):
            raise PyDSParamException(
                "Parameter http_check_condition %s not support.", http_check_condition
            )
        self.http_check_condition = http_check_condition
        if (
            http_check_condition != HttpCheckCondition.STATUS_CODE_DEFAULT
            and condition is None
        ):
            raise PyDSParamException(
                "Parameter condition must provider if http_check_condition not equal to STATUS_CODE_DEFAULT"
            )
        self.condition = condition
        self.connect_timeout = connect_timeout
        self.socket_timeout = socket_timeout



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Module java gateway, contain gateway behavior."""

from typing import Any, Optional

from py4j.java_collections import JavaMap
from py4j.java_gateway import GatewayParameters, JavaGateway

from pydolphinscheduler.constants import JavaGatewayDefault
from pydolphinscheduler.exceptions import PyDSJavaGatewayException


def launch_gateway(
    address: Optional[str] = None,
    port: Optional[int] = None,
    auto_convert: Optional[bool] = True,
) -> JavaGateway:
    """Launch java gateway to pydolphinscheduler.

    TODO Note that automatic conversion makes calling Java methods slightly less efficient because
    in the worst case, Py4J needs to go through all registered converters for all parameters.
    This is why automatic conversion is disabled by default.
    """
    gateway_parameters = GatewayParameters(
        address=address or JavaGatewayDefault.SERVER_ADDRESS,
        port=port or JavaGatewayDefault.SERVER_PORT,
        auto_convert=auto_convert or JavaGatewayDefault.AUTO_CONVERT,
    )
    gateway = JavaGateway(gateway_parameters=gateway_parameters)
    return gateway


def gateway_result_checker(
    result: JavaMap,
    msg_check: Optional[str] = JavaGatewayDefault.RESULT_MESSAGE_SUCCESS,
) -> Any:
    """Check weather java gateway result success or not."""
    if (
        result[JavaGatewayDefault.RESULT_STATUS_KEYWORD].toString()
        != JavaGatewayDefault.RESULT_STATUS_SUCCESS
    ):
        raise PyDSJavaGatewayException("Failed when try to got result for java gateway")
    if (
        msg_check is not None
        and result[JavaGatewayDefault.RESULT_MESSAGE_KEYWORD] != msg_check
    ):
        raise PyDSJavaGatewayException("Get result state not success.")
    return result



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# [start workflow_declare]
"""
A example workflow for task datax.

This example will create a workflow named `task_datax`.
`task_datax` is true workflow define and run task task_datax.
You can create data sources `first_mysql` and `first_mysql` through UI.
It creates a task to synchronize datax from the source database to the target database.
"""

from pydolphinscheduler.core.process_definition import ProcessDefinition
from pydolphinscheduler.tasks.datax import CustomDataX, DataX

# datax json template
JSON_TEMPLATE = {
    "job": {
        "content": [
            {
                "reader": {
                    "name": "mysqlreader",
                    "parameter": {
                        "username": "usr",
                        "password": "pwd",
                        "column": ["id", "name", "code", "description"],
                        "splitPk": "id",
                        "connection": [
                            {
                                "table": ["source_table"],
                                "jdbcUrl": ["jdbc:mysql://127.0.0.1:3306/source_db"],
                            }
                        ],
                    },
                },
                "writer": {
                    "name": "mysqlwriter",
                    "parameter": {
                        "writeMode": "insert",
                        "username": "usr",
                        "password": "pwd",
                        "column": ["id", "name"],
                        "connection": [
                            {
                                "jdbcUrl": "jdbc:mysql://127.0.0.1:3306/target_db",
                                "table": ["target_table"],
                            }
                        ],
                    },
                },
            }
        ],
        "setting": {
            "errorLimit": {"percentage": 0, "record": 0},
            "speed": {"channel": 1, "record": 1000},
        },
    }
}

with ProcessDefinition(
    name="task_datax_example",
    tenant="tenant_exists",
) as pd:
    # This task synchronizes the data in `t_ds_project`
    # of `first_mysql` database to `target_project` of `second_mysql` database.
    # You have to make sure data source named `first_mysql` and `second_mysql` exists
    # in your environment.
    task1 = DataX(
        name="task_datax",
        datasource_name="first_mysql",
        datatarget_name="second_mysql",
        sql="select id, name, code, description from source_table",
        target_table="target_table",
    )

    # You can custom json_template of datax to sync data. This task create a new
    # datax job same as task1, transfer record from `first_mysql` to `second_mysql`
    task2 = CustomDataX(name="task_custom_datax", json=str(JSON_TEMPLATE))
    pd.run()
# [end workflow_declare]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# [start workflow_declare]
r"""
A example workflow for task dependent.

This example will create two workflows named `task_dependent` and `task_dependent_external`.
`task_dependent` is true workflow define and run task dependent, while `task_dependent_external`
define outside workflow and task from dependent.

After this script submit, we would get workflow as below:

task_dependent_external:

task_1
task_2
task_3

task_dependent:

task_dependent(this task dependent on task_dependent_external.task_1 and task_dependent_external.task_2).
"""
from pydolphinscheduler.constants import ProcessDefinitionDefault
from pydolphinscheduler.core.process_definition import ProcessDefinition
from pydolphinscheduler.tasks.dependent import And, Dependent, DependentItem, Or
from pydolphinscheduler.tasks.shell import Shell

with ProcessDefinition(
    name="task_dependent_external",
    tenant="tenant_exists",
) as pd:
    task_1 = Shell(name="task_1", command="echo task 1")
    task_2 = Shell(name="task_2", command="echo task 2")
    task_3 = Shell(name="task_3", command="echo task 3")
    pd.submit()

with ProcessDefinition(
    name="task_dependent_example",
    tenant="tenant_exists",
) as pd:
    task = Dependent(
        name="task_dependent",
        dependence=And(
            Or(
                DependentItem(
                    project_name=ProcessDefinitionDefault.PROJECT,
                    process_definition_name="task_dependent_external",
                    dependent_task_name="task_1",
                ),
                DependentItem(
                    project_name=ProcessDefinitionDefault.PROJECT,
                    process_definition_name="task_dependent_external",
                    dependent_task_name="task_2",
                ),
            )
        ),
    )
    pd.submit()
# [end workflow_declare]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Date util function collections."""

from datetime import datetime

from pydolphinscheduler.constants import Delimiter, Time

LEN_SUPPORT_DATETIME = (
    15,
    19,
)

FMT_SHORT = f"{Time.FMT_SHORT_DATE} {Time.FMT_NO_COLON_TIME}"
FMT_DASH = f"{Time.FMT_DASH_DATE} {Time.FMT_STD_TIME}"
FMT_STD = f"{Time.FMT_STD_DATE} {Time.FMT_STD_TIME}"

MAX_DATETIME = datetime(9999, 12, 31, 23, 59, 59)


def conv_to_schedule(src: datetime) -> str:
    """Convert given datetime to schedule date string."""
    return datetime.strftime(src, FMT_STD)


def conv_from_str(src: str) -> datetime:
    """Convert given string to datetime.

    This function give an ability to convert string to datetime, and for now it could handle
    format like:
    - %Y-%m-%d
    - %Y/%m/%d
    - %Y%m%d
    - %Y-%m-%d %H:%M:%S
    - %Y/%m/%d %H:%M:%S
    - %Y%m%d %H%M%S
    If pattern not like above be given will raise NotImplementedError.
    """
    len_ = len(src)
    if len_ == Time.LEN_SHORT_DATE:
        return datetime.strptime(src, Time.FMT_SHORT_DATE)
    elif len_ == Time.LEN_STD_DATE:
        if Delimiter.BAR in src:
            return datetime.strptime(src, Time.FMT_STD_DATE)
        elif Delimiter.DASH in src:
            return datetime.strptime(src, Time.FMT_DASH_DATE)
        else:
            raise NotImplementedError(
                "%s could not be convert to datetime for now.", src
            )
    elif len_ in LEN_SUPPORT_DATETIME:
        if Delimiter.BAR in src and Delimiter.COLON in src:
            return datetime.strptime(src, FMT_STD)
        elif Delimiter.DASH in src and Delimiter.COLON in src:
            return datetime.strptime(src, FMT_DASH)
        elif (
            Delimiter.DASH not in src
            and Delimiter.BAR not in src
            and Delimiter.COLON not in src
        ):
            return datetime.strptime(src, FMT_SHORT)
        else:
            raise NotImplementedError(
                "%s could not be convert to datetime for now.", src
            )
    else:
        raise NotImplementedError("%s could not be convert to datetime for now.", src)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""DolphinScheduler Task and TaskRelation object."""

import logging
from typing import Dict, List, Optional, Sequence, Set, Tuple, Union

from pydolphinscheduler.constants import (
    Delimiter,
    ProcessDefinitionDefault,
    TaskFlag,
    TaskPriority,
    TaskTimeoutFlag,
)
from pydolphinscheduler.core.base import Base
from pydolphinscheduler.core.process_definition import (
    ProcessDefinition,
    ProcessDefinitionContext,
)
from pydolphinscheduler.java_gateway import launch_gateway


class TaskRelation(Base):
    """TaskRelation object, describe the relation of exactly two tasks."""

    # Add attr `_KEY_ATTR` to overwrite :func:`__eq__`, it is make set
    # `Task.process_definition._task_relations` work correctly.
    _KEY_ATTR = {
        "pre_task_code",
        "post_task_code",
    }

    _DEFINE_ATTR = {
        "pre_task_code",
        "post_task_code",
    }

    _DEFAULT_ATTR = {
        "name": "",
        "preTaskVersion": 1,
        "postTaskVersion": 1,
        "conditionType": 0,
        "conditionParams": {},
    }

    def __init__(
        self,
        pre_task_code: int,
        post_task_code: int,
        name: Optional[str] = None,
    ):
        super().__init__(name)
        self.pre_task_code = pre_task_code
        self.post_task_code = post_task_code

    def __hash__(self):
        return hash(f"{self.pre_task_code} {Delimiter.DIRECTION} {self.post_task_code}")


class Task(Base):
    """Task object, parent class for all exactly task type."""

    _DEFINE_ATTR = {
        "name",
        "code",
        "version",
        "task_type",
        "task_params",
        "description",
        "flag",
        "task_priority",
        "worker_group",
        "delay_time",
        "fail_retry_times",
        "fail_retry_interval",
        "timeout_flag",
        "timeout_notify_strategy",
        "timeout",
    }

    _task_custom_attr: set = set()

    DEFAULT_CONDITION_RESULT = {"successNode": [""], "failedNode": [""]}

    def __init__(
        self,
        name: str,
        task_type: str,
        description: Optional[str] = None,
        flag: Optional[str] = TaskFlag.YES,
        task_priority: Optional[str] = TaskPriority.MEDIUM,
        worker_group: Optional[str] = ProcessDefinitionDefault.WORKER_GROUP,
        delay_time: Optional[int] = 0,
        fail_retry_times: Optional[int] = 0,
        fail_retry_interval: Optional[int] = 1,
        timeout_flag: Optional[int] = TaskTimeoutFlag.CLOSE,
        timeout_notify_strategy: Optional = None,
        timeout: Optional[int] = 0,
        process_definition: Optional[ProcessDefinition] = None,
        local_params: Optional[List] = None,
        resource_list: Optional[List] = None,
        dependence: Optional[Dict] = None,
        wait_start_timeout: Optional[Dict] = None,
        condition_result: Optional[Dict] = None,
    ):

        super().__init__(name, description)
        self.task_type = task_type
        self.flag = flag
        self.task_priority = task_priority
        self.worker_group = worker_group
        self.fail_retry_times = fail_retry_times
        self.fail_retry_interval = fail_retry_interval
        self.delay_time = delay_time
        self.timeout_flag = timeout_flag
        self.timeout_notify_strategy = timeout_notify_strategy
        self.timeout = timeout
        self._process_definition = None
        self.process_definition: ProcessDefinition = (
            process_definition or ProcessDefinitionContext.get()
        )
        self._upstream_task_codes: Set[int] = set()
        self._downstream_task_codes: Set[int] = set()
        self._task_relation: Set[TaskRelation] = set()
        # move attribute code and version after _process_definition and process_definition declare
        self.code, self.version = self.gen_code_and_version()
        # Add task to process definition, maybe we could put into property process_definition latter
        if (
            self.process_definition is not None
            and self.code not in self.process_definition.tasks
        ):
            self.process_definition.add_task(self)
        else:
            logging.warning(
                "Task code %d already in process definition, prohibit re-add task.",
                self.code,
            )

        # Attribute for task param
        self.local_params = local_params or []
        self.resource_list = resource_list or []
        self.dependence = dependence or {}
        self.wait_start_timeout = wait_start_timeout or {}
        self._condition_result = condition_result or self.DEFAULT_CONDITION_RESULT

    @property
    def process_definition(self) -> Optional[ProcessDefinition]:
        """Get attribute process_definition."""
        return self._process_definition

    @process_definition.setter
    def process_definition(self, process_definition: Optional[ProcessDefinition]):
        """Set attribute process_definition."""
        self._process_definition = process_definition

    @property
    def condition_result(self) -> Dict:
        """Get attribute condition_result."""
        return self._condition_result

    @condition_result.setter
    def condition_result(self, condition_result: Optional[Dict]):
        """Set attribute condition_result."""
        self._condition_result = condition_result

    @property
    def task_params(self) -> Optional[Dict]:
        """Get task parameter object.

        Will get result to combine _task_custom_attr and custom_attr.
        """
        custom_attr = {
            "local_params",
            "resource_list",
            "dependence",
            "wait_start_timeout",
            "condition_result",
        }
        custom_attr |= self._task_custom_attr
        return self.get_define_custom(custom_attr=custom_attr)

    def __hash__(self):
        return hash(self.code)

    def __lshift__(self, other: Union["Task", Sequence["Task"]]):
        """Implement Task << Task."""
        self.set_upstream(other)
        return other

    def __rshift__(self, other: Union["Task", Sequence["Task"]]):
        """Implement Task >> Task."""
        self.set_downstream(other)
        return other

    def __rrshift__(self, other: Union["Task", Sequence["Task"]]):
        """Call for Task >> [Task] because list don't have __rshift__ operators."""
        self.__lshift__(other)
        return self

    def __rlshift__(self, other: Union["Task", Sequence["Task"]]):
        """Call for Task << [Task] because list don't have __lshift__ operators."""
        self.__rshift__(other)
        return self

    def _set_deps(
        self, tasks: Union["Task", Sequence["Task"]], upstream: bool = True
    ) -> None:
        """
        Set parameter tasks dependent to current task.

        it is a wrapper for :func:`set_upstream` and :func:`set_downstream`.
        """
        if not isinstance(tasks, Sequence):
            tasks = [tasks]

        for task in tasks:
            if upstream:
                self._upstream_task_codes.add(task.code)
                task._downstream_task_codes.add(self.code)

                if self._process_definition:
                    task_relation = TaskRelation(
                        pre_task_code=task.code,
                        post_task_code=self.code,
                        name=f"{task.name} {Delimiter.DIRECTION} {self.name}",
                    )
                    self.process_definition._task_relations.add(task_relation)
            else:
                self._downstream_task_codes.add(task.code)
                task._upstream_task_codes.add(self.code)

                if self._process_definition:
                    task_relation = TaskRelation(
                        pre_task_code=self.code,
                        post_task_code=task.code,
                        name=f"{self.name} {Delimiter.DIRECTION} {task.name}",
                    )
                    self.process_definition._task_relations.add(task_relation)

    def set_upstream(self, tasks: Union["Task", Sequence["Task"]]) -> None:
        """Set parameter tasks as upstream to current task."""
        self._set_deps(tasks, upstream=True)

    def set_downstream(self, tasks: Union["Task", Sequence["Task"]]) -> None:
        """Set parameter tasks as downstream to current task."""
        self._set_deps(tasks, upstream=False)

    # TODO code should better generate in bulk mode when :ref: processDefinition run submit or start
    def gen_code_and_version(self) -> Tuple:
        """
        Generate task code and version from java gateway.

        If task name do not exists in process definition before, if will generate new code and version id
        equal to 0 by java gateway, otherwise if will return the exists code and version.
        """
        # TODO get code from specific project process definition and task name
        gateway = launch_gateway()
        result = gateway.entry_point.getCodeAndVersion(
            self.process_definition._project, self.name
        )
        # result = gateway.entry_point.genTaskCodeList(DefaultTaskCodeNum.DEFAULT)
        # gateway_result_checker(result)
        return result.get("code"), result.get("version")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Task Python."""

import inspect
import types
from typing import Any

from pydolphinscheduler.constants import TaskType
from pydolphinscheduler.core.task import Task
from pydolphinscheduler.exceptions import PyDSParamException


class Python(Task):
    """Task Python object, declare behavior for Python task to dolphinscheduler."""

    _task_custom_attr = {
        "raw_script",
    }

    def __init__(self, name: str, code: Any, *args, **kwargs):
        super().__init__(name, TaskType.PYTHON, *args, **kwargs)
        self._code = code

    @property
    def raw_script(self) -> str:
        """Get python task define attribute `raw_script`."""
        if isinstance(self._code, str):
            return self._code
        elif isinstance(self._code, types.FunctionType):
            py_function = inspect.getsource(self._code)
            return py_function
        else:
            raise PyDSParamException(
                "Parameter code do not support % for now.", type(self._code)
            )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""DolphinScheduler Worker Group object."""

from typing import Optional

from pydolphinscheduler.core.base_side import BaseSide


class WorkerGroup(BaseSide):
    """DolphinScheduler Worker Group object."""

    def __init__(self, name: str, address: str, description: Optional[str] = None):
        super().__init__(name, description)
        self.address = address



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Task shell."""

from pydolphinscheduler.constants import TaskType
from pydolphinscheduler.core.task import Task


class Shell(Task):
    """Task shell object, declare behavior for shell task to dolphinscheduler.

    :param name: A unique, meaningful string for the shell task.
    :param command: One or more command want to run in this task.

        It could be simply command::

            Shell(name=..., command="echo task shell")

        or maybe same commands trying to do complex task::

            command = '''echo task shell step 1;
            echo task shell step 2;
            echo task shell step 3
            '''

            Shell(name=..., command=command)

    """

    # TODO maybe we could use instance name to replace attribute `name`
    #  which is simplify as `task_shell = Shell(command = "echo 1")` and
    #  task.name assign to `task_shell`

    _task_custom_attr = {
        "raw_script",
    }

    def __init__(self, name: str, command: str, *args, **kwargs):
        super().__init__(name, TaskType.SHELL, *args, **kwargs)
        self.raw_script = command



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Module process definition, core class for workflow define."""

import json
from datetime import datetime
from typing import Any, Dict, List, Optional, Set

from pydolphinscheduler.constants import (
    ProcessDefinitionDefault,
    ProcessDefinitionReleaseState,
    TaskType,
)
from pydolphinscheduler.core.base import Base
from pydolphinscheduler.exceptions import PyDSParamException, PyDSTaskNoFoundException
from pydolphinscheduler.java_gateway import launch_gateway
from pydolphinscheduler.side import Project, Tenant, User
from pydolphinscheduler.utils.date import MAX_DATETIME, conv_from_str, conv_to_schedule


class ProcessDefinitionContext:
    """Class process definition context, use when task get process definition from context expression."""

    _context_managed_process_definition: Optional["ProcessDefinition"] = None

    @classmethod
    def set(cls, pd: "ProcessDefinition") -> None:
        """Set attribute self._context_managed_process_definition."""
        cls._context_managed_process_definition = pd

    @classmethod
    def get(cls) -> Optional["ProcessDefinition"]:
        """Get attribute self._context_managed_process_definition."""
        return cls._context_managed_process_definition

    @classmethod
    def delete(cls) -> None:
        """Delete attribute self._context_managed_process_definition."""
        cls._context_managed_process_definition = None


class ProcessDefinition(Base):
    """process definition object, will define process definition attribute, task, relation.

    TODO: maybe we should rename this class, currently use DS object name.
    """

    # key attribute for identify ProcessDefinition object
    _KEY_ATTR = {
        "name",
        "project",
        "tenant",
        "release_state",
        "param",
    }

    _DEFINE_ATTR = {
        "name",
        "description",
        "_project",
        "_tenant",
        "worker_group",
        "timeout",
        "release_state",
        "param",
        "tasks",
        "task_definition_json",
        "task_relation_json",
    }

    def __init__(
        self,
        name: str,
        description: Optional[str] = None,
        schedule: Optional[str] = None,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        timezone: Optional[str] = ProcessDefinitionDefault.TIME_ZONE,
        user: Optional[str] = ProcessDefinitionDefault.USER,
        project: Optional[str] = ProcessDefinitionDefault.PROJECT,
        tenant: Optional[str] = ProcessDefinitionDefault.TENANT,
        queue: Optional[str] = ProcessDefinitionDefault.QUEUE,
        worker_group: Optional[str] = ProcessDefinitionDefault.WORKER_GROUP,
        timeout: Optional[int] = 0,
        release_state: Optional[str] = ProcessDefinitionReleaseState.ONLINE,
        param: Optional[Dict] = None,
    ):
        super().__init__(name, description)
        self.schedule = schedule
        self._start_time = start_time
        self._end_time = end_time
        self.timezone = timezone
        self._user = user
        self._project = project
        self._tenant = tenant
        self._queue = queue
        self.worker_group = worker_group
        self.timeout = timeout
        self.release_state = release_state
        self.param = param
        self.tasks: dict = {}
        # TODO how to fix circle import
        self._task_relations: set["TaskRelation"] = set()  # noqa: F821
        self._process_definition_code = None

    def __enter__(self) -> "ProcessDefinition":
        ProcessDefinitionContext.set(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        ProcessDefinitionContext.delete()

    @property
    def tenant(self) -> Tenant:
        """Get attribute tenant."""
        return Tenant(self._tenant)

    @tenant.setter
    def tenant(self, tenant: Tenant) -> None:
        """Set attribute tenant."""
        self._tenant = tenant.name

    @property
    def project(self) -> Project:
        """Get attribute project."""
        return Project(self._project)

    @project.setter
    def project(self, project: Project) -> None:
        """Set attribute project."""
        self._project = project.name

    @property
    def user(self) -> User:
        """Get user object.

        For now we just get from python side but not from java gateway side, so it may not correct.
        """
        return User(
            self._user,
            ProcessDefinitionDefault.USER_PWD,
            ProcessDefinitionDefault.USER_EMAIL,
            ProcessDefinitionDefault.USER_PHONE,
            self._tenant,
            self._queue,
            ProcessDefinitionDefault.USER_STATE,
        )

    @staticmethod
    def _parse_datetime(val: Any) -> Any:
        if val is None or isinstance(val, datetime):
            return val
        elif isinstance(val, str):
            return conv_from_str(val)
        else:
            raise PyDSParamException("Do not support value type %s for now", type(val))

    @property
    def start_time(self) -> Any:
        """Get attribute start_time."""
        return self._parse_datetime(self._start_time)

    @start_time.setter
    def start_time(self, val) -> None:
        """Set attribute start_time."""
        self._start_time = val

    @property
    def end_time(self) -> Any:
        """Get attribute end_time."""
        return self._parse_datetime(self._end_time)

    @end_time.setter
    def end_time(self, val) -> None:
        """Set attribute end_time."""
        self._end_time = val

    @property
    def param_json(self) -> Optional[List[Dict]]:
        """Return param json base on self.param."""
        # Handle empty dict and None value
        if not self.param:
            return []
        return [
            {
                "prop": k,
                "direct": "IN",
                "type": "VARCHAR",
                "value": v,
            }
            for k, v in self.param.items()
        ]

    @property
    def task_definition_json(self) -> List[Dict]:
        """Return all tasks definition in list of dict."""
        if not self.tasks:
            return [self.tasks]
        else:
            return [task.get_define() for task in self.tasks.values()]

    @property
    def task_relation_json(self) -> List[Dict]:
        """Return all relation between tasks pair in list of dict."""
        if not self.tasks:
            return [self.tasks]
        else:
            self._handle_root_relation()
            return [tr.get_define() for tr in self._task_relations]

    @property
    def schedule_json(self) -> Optional[Dict]:
        """Get schedule parameter json object. This is requests from java gateway interface."""
        if not self.schedule:
            return None
        else:
            start_time = conv_to_schedule(
                self.start_time if self.start_time else datetime.now()
            )
            end_time = conv_to_schedule(
                self.end_time if self.end_time else MAX_DATETIME
            )
            return {
                "startTime": start_time,
                "endTime": end_time,
                "crontab": self.schedule,
                "timezoneId": self.timezone,
            }

    # TODO inti DAG's tasks are in the same location with default {x: 0, y: 0}
    @property
    def task_location(self) -> List[Dict]:
        """Return all tasks location for all process definition.

        For now, we only set all location with same x and y valued equal to 0. Because we do not
        find a good way to set task locations. This is requests from java gateway interface.
        """
        if not self.tasks:
            return [self.tasks]
        else:
            return [{"taskCode": task_code, "x": 0, "y": 0} for task_code in self.tasks]

    @property
    def task_list(self) -> List["Task"]:  # noqa: F821
        """Return list of tasks objects."""
        return list(self.tasks.values())

    def _handle_root_relation(self):
        """Handle root task property :class:`pydolphinscheduler.core.task.TaskRelation`.

        Root task in DAG do not have dominant upstream node, but we have to add an exactly default
        upstream task with task_code equal to `0`. This is requests from java gateway interface.
        """
        from pydolphinscheduler.core.task import TaskRelation

        post_relation_code = set()
        for relation in self._task_relations:
            post_relation_code.add(relation.post_task_code)
        for task in self.task_list:
            if task.code not in post_relation_code:
                root_relation = TaskRelation(pre_task_code=0, post_task_code=task.code)
                self._task_relations.add(root_relation)

    def add_task(self, task: "Task") -> None:  # noqa: F821
        """Add a single task to process definition."""
        self.tasks[task.code] = task
        task._process_definition = self

    def add_tasks(self, tasks: List["Task"]) -> None:  # noqa: F821
        """Add task sequence to process definition, it a wrapper of :func:`add_task`."""
        for task in tasks:
            self.add_task(task)

    def get_task(self, code: str) -> "Task":  # noqa: F821
        """Get task object from process definition by given code."""
        if code not in self.tasks:
            raise PyDSTaskNoFoundException(
                "Task with code %s can not found in process definition %",
                (code, self.name),
            )
        return self.tasks[code]

    # TODO which tying should return in this case
    def get_tasks_by_name(self, name: str) -> Set["Task"]:  # noqa: F821
        """Get tasks object by given name, if will return all tasks with this name."""
        find = set()
        for task in self.tasks.values():
            if task.name == name:
                find.add(task)
        return find

    def get_one_task_by_name(self, name: str) -> "Task":  # noqa: F821
        """Get exact one task from process definition by given name.

        Function always return one task even though this process definition have more than one task with
        this name.
        """
        tasks = self.get_tasks_by_name(name)
        if not tasks:
            raise PyDSTaskNoFoundException(f"Can not find task with name {name}.")
        return tasks.pop()

    def run(self):
        """Submit and Start ProcessDefinition instance.

        Shortcut for function :func:`submit` and function :func:`start`. Only support manual start workflow
        for now, and schedule run will coming soon.
        :return:
        """
        self.submit()
        self.start()

    def _ensure_side_model_exists(self):
        """Ensure process definition side model exists.

        For now, side object including :class:`pydolphinscheduler.side.project.Project`,
        :class:`pydolphinscheduler.side.tenant.Tenant`, :class:`pydolphinscheduler.side.user.User`.
        If these model not exists, would create default value in
        :class:`pydolphinscheduler.constants.ProcessDefinitionDefault`.
        """
        # TODO used metaclass for more pythonic
        self.tenant.create_if_not_exists(self._queue)
        # model User have to create after Tenant created
        self.user.create_if_not_exists()
        # Project model need User object exists
        self.project.create_if_not_exists(self._user)

    def _pre_submit_check(self):
        """Check specific condition satisfy before.

        This method should be called before process definition submit to java gateway
        For now, we have below checker:
        * `self.param` should be set if task `switch` in this workflow.
        """
        if (
            any([task.task_type == TaskType.SWITCH for task in self.tasks.values()])
            and self.param is None
        ):
            raise PyDSParamException(
                "Parameter param must be provider if task Switch in process definition."
            )

    def submit(self) -> int:
        """Submit ProcessDefinition instance to java gateway."""
        self._ensure_side_model_exists()
        self._pre_submit_check()

        gateway = launch_gateway()
        self._process_definition_code = gateway.entry_point.createOrUpdateProcessDefinition(
            self._user,
            self._project,
            self.name,
            str(self.description) if self.description else "",
            json.dumps(self.param_json),
            json.dumps(self.schedule_json) if self.schedule_json else None,
            json.dumps(self.task_location),
            self.timeout,
            self.worker_group,
            self._tenant,
            # TODO add serialization function
            json.dumps(self.task_relation_json),
            json.dumps(self.task_definition_json),
            None,
        )
        return self._process_definition_code

    def start(self) -> None:
        """Create and start ProcessDefinition instance.

        which post to `start-process-instance` to java gateway
        """
        gateway = launch_gateway()
        gateway.entry_point.execProcessInstance(
            self._user,
            self._project,
            self.name,
            "",
            self.worker_group,
            24 * 3600,
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""DolphinScheduler User object."""

from typing import Optional

from pydolphinscheduler.core.base_side import BaseSide
from pydolphinscheduler.java_gateway import launch_gateway


class User(BaseSide):
    """DolphinScheduler User object."""

    _KEY_ATTR = {
        "name",
        "password",
        "email",
        "phone",
        "tenant",
        "queue",
        "status",
    }

    def __init__(
        self,
        name: str,
        password: str,
        email: str,
        phone: str,
        tenant: str,
        queue: Optional[str] = None,
        status: Optional[int] = 1,
    ):
        super().__init__(name)
        self.password = password
        self.email = email
        self.phone = phone
        self.tenant = tenant
        self.queue = queue
        self.status = status

    def create_if_not_exists(self, **kwargs):
        """Create User if not exists."""
        gateway = launch_gateway()
        gateway.entry_point.createUser(
            self.name,
            self.password,
            self.email,
            self.phone,
            self.tenant,
            self.queue,
            self.status,
        )
        # TODO recover result checker
        # gateway_result_checker(result, None)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""The script for setting up pydolphinscheduler."""
import logging
import os
import sys
from distutils.dir_util import remove_tree
from os.path import dirname, join
from typing import List

from setuptools import Command, find_packages, setup

if sys.version_info[0] < 3:
    raise Exception(
        "pydolphinscheduler does not support Python 2. Please upgrade to Python 3."
    )

logger = logging.getLogger(__name__)

version = "2.0.7"

# Start package required
prod = [
    "py4j~=0.10",
]

build = [
    "build",
    "setuptools>=42",
    "wheel",
]

doc = [
    "sphinx>=4.3",
    "sphinx_rtd_theme>=1.0",
]

test = [
    "pytest>=6.2",
    "freezegun>=1.1",
    "coverage>=6.1",
]

style = [
    "flake8>=4.0",
    "flake8-docstrings>=1.6",
    "flake8-black>=0.2",
    "isort>=5.10",
]

dev = style + test + doc + build

all_dep = prod + dev
# End package required


def read(*names, **kwargs):
    """Read file content from given file path."""
    return open(
        join(dirname(__file__), *names), encoding=kwargs.get("encoding", "utf8")
    ).read()


class CleanCommand(Command):
    """Command to clean up python api before setup by running `python setup.py pre_clean`."""

    description = "Clean up project root"
    user_options: List[str] = []
    clean_list = [
        "build",
        "htmlcov",
        "dist",
        ".pytest_cache",
        ".coverage",
    ]

    def initialize_options(self) -> None:
        """Set default values for options."""
        pass

    def finalize_options(self) -> None:
        """Set final values for options."""
        pass

    def run(self) -> None:
        """Run and remove temporary files."""
        for cl in self.clean_list:
            if not os.path.exists(cl):
                logger.info("Path %s do not exists.", cl)
            elif os.path.isdir(cl):
                remove_tree(cl)
            else:
                os.remove(cl)
        logger.info("Finish pre_clean process.")


setup(
    name="apache-dolphinscheduler",
    version=version,
    license="Apache License 2.0",
    description="Apache DolphinScheduler Python API",
    long_description=read("README.md"),
    # Make sure pypi is expecting markdown
    long_description_content_type="text/markdown",
    author="Apache Software Foundation",
    author_email="dev@dolphinscheduler.apache.org",
    url="https://dolphinscheduler.apache.org/",
    python_requires=">=3.6",
    keywords=[
        "dolphinscheduler",
        "workflow",
        "scheduler",
        "taskflow",
    ],
    project_urls={
        "Homepage": "https://dolphinscheduler.apache.org",
        "Documentation": "https://dolphinscheduler.apache.org/python/index.html",
        "Source": "https://github.com/apache/dolphinscheduler/dolphinscheduler-python/pydolphinscheduler",
        "Issue Tracker": "https://github.com/apache/dolphinscheduler/issues",
        "Discussion": "https://github.com/apache/dolphinscheduler/discussions",
        "Twitter": "https://twitter.com/dolphinschedule",
    },
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    package_data={
        "examples": ["examples.tutorial.py"],
    },
    platforms=["any"],
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: User Interfaces",
    ],
    install_requires=prod,
    extras_require={
        "all": all_dep,
        "dev": dev,
        "style": style,
        "test": test,
        "doc": doc,
        "build": build,
    },
    cmdclass={
        "pre_clean": CleanCommand,
    },
)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Constants for pydolphinscheduler."""


class ProcessDefinitionReleaseState:
    """Constants for :class:`pydolphinscheduler.core.process_definition.ProcessDefinition` release state."""

    ONLINE: str = "ONLINE"
    OFFLINE: str = "OFFLINE"


class ProcessDefinitionDefault:
    """Constants default value for :class:`pydolphinscheduler.core.process_definition.ProcessDefinition`."""

    PROJECT: str = "project-pydolphin"
    TENANT: str = "tenant_pydolphin"
    USER: str = "userPythonGateway"
    # TODO simple set password same as username
    USER_PWD: str = "userPythonGateway"
    USER_EMAIL: str = "userPythonGateway@dolphinscheduler.com"
    USER_PHONE: str = "11111111111"
    USER_STATE: int = 1
    QUEUE: str = "queuePythonGateway"
    WORKER_GROUP: str = "default"
    TIME_ZONE: str = "Asia/Shanghai"


class TaskPriority(str):
    """Constants for task priority."""

    HIGHEST = "HIGHEST"
    HIGH = "HIGH"
    MEDIUM = "MEDIUM"
    LOW = "LOW"
    LOWEST = "LOWEST"


class TaskFlag(str):
    """Constants for task flag."""

    YES = "YES"
    NO = "NO"


class TaskTimeoutFlag(str):
    """Constants for task timeout flag."""

    CLOSE = "CLOSE"


class TaskType(str):
    """Constants for task type, it will also show you which kind we support up to now."""

    SHELL = "SHELL"
    HTTP = "HTTP"
    PYTHON = "PYTHON"
    SQL = "SQL"
    SUB_PROCESS = "SUB_PROCESS"
    PROCEDURE = "PROCEDURE"
    DATAX = "DATAX"
    DEPENDENT = "DEPENDENT"
    CONDITIONS = "CONDITIONS"
    SWITCH = "SWITCH"
    FLINK = "FLINK"
    SPARK = "SPARK"
    MR = "MR"


class DefaultTaskCodeNum(str):
    """Constants and default value for default task code number."""

    DEFAULT = 1


class JavaGatewayDefault(str):
    """Constants and default value for java gateway."""

    RESULT_MESSAGE_KEYWORD = "msg"
    RESULT_MESSAGE_SUCCESS = "success"

    RESULT_STATUS_KEYWORD = "status"
    RESULT_STATUS_SUCCESS = "SUCCESS"

    RESULT_DATA = "data"

    SERVER_ADDRESS = "127.0.0.1"
    SERVER_PORT = 25333
    AUTO_CONVERT = True


class Delimiter(str):
    """Constants for delimiter."""

    BAR = "-"
    DASH = "/"
    COLON = ":"
    UNDERSCORE = "_"
    DIRECTION = "->"


class Time(str):
    """Constants for date."""

    FMT_STD_DATE = "%Y-%m-%d"
    LEN_STD_DATE = 10

    FMT_DASH_DATE = "%Y/%m/%d"

    FMT_SHORT_DATE = "%Y%m%d"
    LEN_SHORT_DATE = 8

    FMT_STD_TIME = "%H:%M:%S"
    FMT_NO_COLON_TIME = "%H%M%S"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Init pydolphinscheduler.tasks package."""

from pydolphinscheduler.tasks.condition import FAILURE, SUCCESS, And, Condition, Or
from pydolphinscheduler.tasks.datax import CustomDataX, DataX
from pydolphinscheduler.tasks.dependent import Dependent
from pydolphinscheduler.tasks.flink import Flink
from pydolphinscheduler.tasks.http import Http
from pydolphinscheduler.tasks.map_reduce import MR
from pydolphinscheduler.tasks.procedure import Procedure
from pydolphinscheduler.tasks.python import Python
from pydolphinscheduler.tasks.shell import Shell
from pydolphinscheduler.tasks.spark import Spark
from pydolphinscheduler.tasks.sql import Sql
from pydolphinscheduler.tasks.sub_process import SubProcess
from pydolphinscheduler.tasks.switch import Branch, Default, Switch, SwitchCondition

__all__ = [
    "Condition",
    "DataX",
    "Dependent",
    "Flink",
    "Http",
    "MR",
    "Procedure",
    "Python",
    "Shell",
    "Spark",
    "Sql",
    "SubProcess",
    "Switch",
]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""DolphinScheduler Tenant object."""

from typing import Optional

from pydolphinscheduler.constants import ProcessDefinitionDefault
from pydolphinscheduler.core.base_side import BaseSide
from pydolphinscheduler.java_gateway import launch_gateway


class Tenant(BaseSide):
    """DolphinScheduler Tenant object."""

    def __init__(
        self,
        name: str = ProcessDefinitionDefault.TENANT,
        queue: str = ProcessDefinitionDefault.QUEUE,
        description: Optional[str] = None,
    ):
        super().__init__(name, description)
        self.queue = queue

    def create_if_not_exists(
        self, queue_name: str, user=ProcessDefinitionDefault.USER
    ) -> None:
        """Create Tenant if not exists."""
        gateway = launch_gateway()
        gateway.entry_point.createTenant(self.name, self.description, queue_name)
        # gateway_result_checker(result, None)
